"""All the process that can be run using nox.

The nox run are build in isolated environment that will be stored in .nox. to force the venv update, remove the .nox/xxx folder.
"""

import nox
import tomli


@nox.session(reuse_venv=True)
def lint(session):
    """Apply the pre-commits."""
    session.install("pre-commit")
    session.run("pre-commit", "run", "--a", *session.posargs)


@nox.session(reuse_venv=True)
def app(session):
    """Run the application."""
    init_notebook = tomli.load("pyproject.toml")["sepal-ui"]["init-notebook"]
    session.install("-r", "requirements.txt")
    session.run("jupyter", "trust", init_notebook)
    session.run("voila", "--debug", init_notebook)


@nox.session(reuse_venv=True)
def jupyter(session):
    """Run the application in a jupyter notebook."""
    session.install("-r", "requirements.txt")
    session.run("jupyter", "trust", "no_ui.ipynb")
    session.run("jupyter", "notebook", "no_ui.ipynb")



"""Helpers to customize the display of sepal-ui widgets and maps."""

from pathlib import Path
from types import SimpleNamespace
from typing import Tuple

import ipyvuetify as v
from IPython.display import HTML, Javascript, display
from ipyvuetify._version import semver
from ipywidgets import Widget
from traitlets import Bool, HasTraits, Unicode, link

import sepal_ui.scripts.utils as su
from sepal_ui.conf import config

################################################################################
# access the folders where style information is stored (layers, widgets)
#

# the colors are set using tables as follow.
# 1 (True): dark theme
# 0 (false): light theme
JSON_DIR: Path = Path(__file__).parent / "json"
"The path to the json style folder"

CSS_DIR: Path = Path(__file__).parent / "css"
"The path to the css style folder"

JS_DIR: Path = Path(__file__).parent / "js"
"The path to the js style folder"

################################################################################
# define all the colors that we want to use in the theme
#


TYPES: Tuple[str, ...] = (
    "info",
    "primary",
    "primary_contarst",
    "secondary",
    "secondary_contrast",
    "accent",
    "error",
    "success",
    "warning",
    "anchor",
    "main",
    "darker",
    "bg",
    "menu",
)
"The different types defined by ipyvuetify"


class ThemeColors(Widget):

    _model_name = Unicode("ThemeColorsModel").tag(sync=True)

    _model_module = Unicode("jupyter-vuetify").tag(sync=True)

    _view_module_version = Unicode(semver).tag(sync=True)

    _model_module_version = Unicode(semver).tag(sync=True)

    _theme_name = Unicode().tag(sync=True)

    primary = Unicode().tag(sync=True)
    primary_contrast = Unicode().tag(sync=True)
    secondary = Unicode().tag(sync=True)
    secondary_contrast = Unicode().tag(sync=True)
    accent = Unicode().tag(sync=True)
    error = Unicode().tag(sync=True)
    info = Unicode().tag(sync=True)
    success = Unicode().tag(sync=True)
    warning = Unicode().tag(sync=True)
    anchor = Unicode(None, allow_none=True).tag(sync=True)
    main = Unicode().tag(sync=True)
    bg = Unicode().tag(sync=True)
    menu = Unicode().tag(sync=True)
    darker = Unicode().tag(sync=True)


dark_theme_colors = ThemeColors(
    _theme_name="dark",
    primary="#76591e",
    primary_contrast="#bf8f2d",  # a bit lighter than the primary color
    secondary="#363e4f",
    secondary_contrast="#5d76ab",
    error="#a63228",
    info="#c5c6c9",
    success="#3f802a",
    warning="#b8721d",
    accent="#272727",
    anchor="#f3f3f3",
    main="#24221f",
    darker="#1a1a1a",
    bg="#121212",
    menu="#424242",
)

light_theme_colors = ThemeColors(
    _theme_name="light",
    primary="#5BB624",
    primary_contrast="#76b353",
    accent="#f3f3f3",
    anchor="#f3f3f3",
    secondary="#2199C4",
    secondary_contrast="#5d76ab",
    success=v.theme.themes.light.success,
    info=v.theme.themes.light.info,
    warning=v.theme.themes.light.warning,
    error=v.theme.themes.light.error,
    main="#2196f3",  # used by appbar and versioncard
    darker="#ffffff",  # used for the navdrawer
    bg="#FFFFFF",
    menu="#FFFFFF",
)

DARK_THEME = {k: v for k, v in dark_theme_colors.__dict__["_trait_values"].items() if k in TYPES}
"colors used for the dark theme"

LIGHT_THEME = {k: v for k, v in light_theme_colors.__dict__["_trait_values"].items() if k in TYPES}
"colors used for the light theme"


# override the default theme with the custom ones
v.theme.themes.light = light_theme_colors
v.theme.themes.dark = dark_theme_colors

################################################################################
# define classes and method to make the application responsive
#


def get_theme() -> str:
    """Get theme name from the config file (default to dark).

    Returns:
        The theme to use
    """
    return config.get("sepal-ui", "theme", fallback="dark")


class SepalColor(HasTraits, SimpleNamespace):

    _dark_theme: Bool = Bool(True if get_theme() == "dark" else False).tag(sync=True)
    "Whether to use dark theme or not. By changing this value, the theme value will be stored in the conf file. Is only intended to be accessed in development mode."

    def __init__(self) -> None:
        """Custom simple name space to store and access to the sepal_ui colors and with a magic method to display theme."""
        link((self, "_dark_theme"), (v.theme, "dark"))
        v.theme.observe(lambda *x: self.set_colors(), "dark")

        self.set_colors()

    def set_colors(self) -> None:
        """Set the current hexadecimal color in the object."""
        # Get get current theme name
        self.theme_name = "dark" if self._dark_theme else "light"

        # Save "new" theme in configuration file
        su.set_config("theme", self.theme_name)

        self.colors_dict = DARK_THEME if self._dark_theme else LIGHT_THEME
        SimpleNamespace.__init__(self, **self.colors_dict)

    def _repr_html_(self, *_) -> str:
        """Rich display of the color palette in an HTML frontend."""
        s = 60
        html = f"<h3>Current theme: {self.theme_name}</h3><table>"
        items = {k: v for k, v in self.colors_dict.items()}.items()

        for name, color in items:
            c = su.to_colors(color)
            html += f"""
            <th>
                <svg width='{s}' height='{s}'>
                    <rect width='{s}' height='{s}' style='fill:{c};
                    stroke-width:1;stroke:rgb(255,255,255)'/>
                </svg>
            </th>
            """

        html += "</tr><tr>"
        html += "".join([f"<td>{name}</br>{color}</td>" for name, color in items])
        html += "</tr></table>"

        return html


# load custom styling of sepal_ui
sepal_ui_css = HTML(f"<style>{(CSS_DIR / 'custom.css').read_text()}</style>")

# load fa-6
fa_css = HTML(
    '<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.2.1/css/all.min.css"/>'
)

# create a small hack to remove fontawesome from the html output
clean_fa_js = Javascript((JS_DIR / "fontawesome.js").read_text())

# display all
display(sepal_ui_css, fa_css, clean_fa_js)



"""All directories used in the application."""

from pathlib import Path

# this directory is the root directory of all sepal dashboard app.
# Change it if you develop in another environment.
module_dir = Path.home() / "module_results"
module_dir.mkdir(exist_ok=True)



"""Custom implementation of the marker cluster to hide it at once."""

from ipyleaflet import MarkerCluster
from traitlets import Bool, observe


class MarkerCluster(MarkerCluster):
    """Overwrite the MarkerCluster to hide all the underlying cluster at once.

    .. todo::
        remove when https://github.com/jupyter-widgets/ipyleaflet/issues/1108 is solved
    """

    visible = Bool(True).tag(sync=True)

    @observe("visible")
    def toggle_markers(self, change):
        """change the marker value according to the cluster viz."""
        for marker in self.markers:
            marker.visible = self.visible



"""Module to load basemaps from different providers."""

from typing import Optional

from box import Box
from ipyleaflet import TileLayer
from xyzservices import TileProvider
from xyzservices import providers as xyz

xyz_tiles: dict = {
    "OpenStreetMap": {
        "url": "https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png",
        "attribution": "OpenStreetMap",
        "name": "OpenStreetMap",
    },
    "ROADMAP": {
        "url": "https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Maps",
    },
    "SATELLITE": {
        "url": "https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Satellite",
    },
    "TERRAIN": {
        "url": "https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Terrain",
    },
    "HYBRID": {
        "url": "https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Satellite",
    },
}
"Custom XYZ tile services."


def get_xyz_dict(
    free_only: bool = True,
    _collection: Optional[dict] = None,
    _output: Optional[dict] = None,
) -> dict:
    """Returns a dictionary of xyz services.

    Args:
        free_only: Whether to return only free xyz tile services that do not require an access token.
        _collection: the collection to anylize (subset of :code:`xyz`)
        _output: the dict to use as an output (mutable object)

    Returns:
        A dictionary of xyz services.
    """
    # the 2 following lies avoid to display xyz descriptor in the method documentation
    # do not replace in the prototype default values
    _collection = xyz if _collection is None else _collection
    _output = {} if _output is None else _output

    for v in _collection.values():
        if isinstance(v, TileProvider):
            if not (v.requires_token() and free_only):
                _output[v.name] = v
        else:  # it's a Bunch
            get_xyz_dict(free_only, v, _output)

    return _output


def xyz_to_leaflet() -> dict:
    """Convert all available xyz tile services to ipyleaflet tile layers.

    Adapted from https://github.com/giswqs/geemap.

    Returns:
        A dictionary of ipyleaflet tile layers.
    """
    leaflet_dict = {}

    for key in xyz_tiles:
        name = xyz_tiles[key]["name"]
        url = xyz_tiles[key]["url"]
        attribution = xyz_tiles[key]["attribution"]
        leaflet_dict[key] = TileLayer(
            url=url, name=name, attribution=attribution, max_zoom=22, base=True
        )

    for item in get_xyz_dict().values():
        leaflet_dict[item.name] = TileLayer(
            url=item.build_url(),
            name=item.name,
            max_zoom=item.get("max_zoom", 22),
            attribution=item.attribution,
            base=True,
        )

    return leaflet_dict


basemap_tiles: Box = Box(xyz_to_leaflet(), frozen_box=True)
"the basemaps list as a box"



"""Module to gather all scripts of the application.

If you only have few widgets, a module is not necessary and you can simply use a scripts.py file
In a big module with lot of custom scripts, it can make sense to split things in separate files for the sake of maintenance

If you use a module import all the functions here to only have 1 call to make
"""



"""A demo tile showing some of the most frequent widgets available in the sepal_ui library."""

import ipyvuetify as v
import sepal_ui.sepalwidgets as sw
from sepal_ui.mapping.sepal_map import SepalMap


class WidgetTile(sw.Card):
    def __init__(self):
        """A demo tile showing all the widgets available in the sepal_ui library."""
        # create the card
        super().__init__(row=True, wrap=True, _metadata={"mount_id": "widget_tile"})

        # Create widgets

        select = sw.Select(
            label="Select",
            v_model="option1",
            items=["option1", "option2", "option3"],
            outlined=True,
        )

        slider = sw.Slider(
            label="Slider",
            v_model=50,
            max=100,
            min=0,
            thumb_label=True,
            step=1,
        )

        btn = sw.Btn("Button", color="primary", outlined=True)
        btn1 = sw.Btn("Button", color="secondary", outlined=True)
        btn2 = sw.Btn("Button", color="primary")
        btn3 = sw.Btn("Button", color="secondary")
        file_input = sw.FileInput()
        map_ = SepalMap(zoom=10)

        title = v.CardTitle(children=["test widgets tile"])

        card = v.Card(class_="mb-4 pa-2", raised=True, xs12=True, children=[title, map_])

        loading_card = v.Card(
            class_="mb-6 pa-2",
            raised=True,
            xs12=True,
            children=[title] + ["content " * 100],
            loading=True,
        )

        warning_alert = sw.Alert().add_msg("This is a warning alert", type_="warning")
        error_alert = sw.Alert().add_msg("This is an error alert", type_="error")
        info_alert = sw.Alert().add_msg("This is an info alert", type_="info")
        success_alert = sw.Alert().add_msg("This is a success alert", type_="success")

        # Add widgets to the layout

        self.children = [
            card,
            v.Flex(class_="mb-2", xs12=True, lg6=True, xl4=True, children=[select]),
            v.Flex(class_="mb-2", xs12=True, lg6=True, xl4=True, children=[file_input]),
            v.Flex(class_="mb-2", xs12=True, lg6=True, xl4=True, children=[slider]),
            v.Flex(class_="mb-2", xs12=True, lg6=True, xl4=True, children=[btn]),
            v.Flex(class_="mb-2", xs12=True, lg6=True, xl4=True, children=[btn1]),
            v.Flex(class_="mb-2", xs12=True, lg6=True, xl4=True, children=[btn2]),
            v.Flex(class_="mb-2", xs12=True, lg6=True, xl4=True, children=[btn3]),
            v.Flex(
                class_="mb-2",
                xs12=True,
                lg6=True,
                xl4=True,
                children=[warning_alert],
            ),
            v.Flex(
                class_="mb-2",
                xs12=True,
                lg6=True,
                xl4=True,
                children=[error_alert],
            ),
            v.Flex(
                class_="mb-2",
                xs12=True,
                lg6=True,
                xl4=True,
                children=[info_alert],
            ),
            v.Flex(
                class_="mb-2",
                xs12=True,
                lg6=True,
                xl4=True,
                children=[success_alert],
            ),
            loading_card,
        ]



"""All the heleper methods to interface Google Earthengine with sepal-ui."""

import time
from pathlib import Path
from typing import List, Union

import ee
import ipyvuetify as v

from sepal_ui.message import ms
from sepal_ui.scripts import decorator as sd


@sd.need_ee
def wait_for_completion(task_descripsion: str, widget_alert: v.Alert = None) -> str:
    """Wait until the selected process is finished. Display some output information.

    Args:
        task_descripsion: name of the running task
        widget_alert: alert to display the output messages

    Returns:
        the final state of the task
    """
    state = "UNSUBMITTED"
    while state != "COMPLETED":

        # print in a widget
        if widget_alert:
            widget_alert.add_live_msg(ms.status.format(state))

        # wait 5 seconds
        time.sleep(5)

        # search for the task in task_list
        current_task = is_task(task_descripsion)
        state = current_task.state

        if state == "FAILED":
            raise Exception(ms.status.format(state))

    # print in a widget
    if widget_alert:
        widget_alert.add_live_msg(ms.status.format(state), "success")

    return state


@sd.need_ee
def is_task(task_descripsion: str) -> ee.batch.Task:
    """Search for the described task in the user Task list return None if nothing is found.

    Args:
        task_descripsion: the task description

    Returns:
        return the found task else None
    """
    current_task = None
    for task in ee.batch.Task.list():
        if task.config["description"] == task_descripsion:
            current_task = task
            break

    return current_task


@sd.need_ee
def is_running(task_descripsion: str) -> ee.batch.Task:
    """Search for the described task in the user Task list return None if nothing is currently running.

    Args:
        task_descripsion: the task description

    Returns:
        return the found task else None
    """
    current_task = is_task(task_descripsion)
    if current_task:
        if current_task.state not in ["RUNNING", "READY"]:
            current_task = None

    return current_task


@sd.need_ee
def get_assets(folder: Union[str, Path] = "") -> List[dict]:
    """Get all the assets from the parameter folder. every nested asset will be displayed.

    Args:
        folder: the initial GEE folder

    Returns:
        the asset list. each asset is a dict with 3 keys: 'type', 'name' and 'id'
    """
    # set the folder and init the list
    asset_list = []
    folder = str(folder) or f"projects/{ee.data._cloud_api_user_project}/assets/"

    def _recursive_get(folder, asset_list):

        # loop in the assets
        for asset in ee.data.listAssets({"parent": folder})["assets"]:
            asset_list += [asset]
            if asset["type"] == "FOLDER":
                asset_list = _recursive_get(asset["name"], asset_list)

        return asset_list

    return _recursive_get(folder, asset_list)


@sd.need_ee
def is_asset(asset_name: str, folder: Union[str, Path] = "") -> bool:
    """Check if the asset already exist in the user asset folder.

    Args:
        asset_descripsion: the descripsion of the asset
        folder: the folder of the glad assets

    Returns:
        true if already in folder
    """
    # get the folder
    folder = str(folder) or f"projects/{ee.data._cloud_api_user_project}/assets/"

    # get all the assets
    asset_list = get_assets(folder)

    # search for asset existence
    exist = False
    for asset in asset_list:
        if asset_name == asset["name"]:
            exist = True
            break

    return exist


@sd.need_ee
def delete_assets(asset_id: str, dry_run: bool = True) -> None:
    """Delete the selected asset and all its content.

    This method will delete all the files and folders existing in an asset folder. By default a dry run will be launched and if you are satisfyed with the displayed names, change the ``dry_run`` variable to ``False``. No other warnng will be displayed.

    .. warning::

        If this method is used on the root directory you will loose all your data, it's highly recommended to use a dry run first and carefully review the destroyed files.

    Args:
        asset_id: the Id of the asset or a folder
        dry_run: whether or not a dry run should be launched. dry run will only display the files name without deleting them.
    """
    # define the action to execute for each asset based on the dry run mode
    def delete(id: str) -> None:
        if dry_run is True:
            print(f"to be deleted: {id}")
        else:
            print(f"deleting: {id}")
            ee.data.deleteAsset(id)

        return

    # identify the type of asset
    asset_info = ee.data.getAsset(asset_id)

    if asset_info["type"] == "FOLDER":

        # get all the assets
        asset_list = get_assets(folder=asset_id)

        # split the files by nesting levels
        # we will need to delete the more nested files first
        assets_ordered = {}
        for asset in asset_list:
            lvl = len(asset["id"].split("/"))
            assets_ordered.setdefault(lvl, [])
            assets_ordered[lvl].append(asset)

        # delete all items starting from the more nested one but not folders
        assets_ordered = dict(sorted(assets_ordered.items(), reverse=True))
        for lvl in assets_ordered:
            for i in assets_ordered[lvl]:
                delete(i["name"])

    # delete the initial folder/asset
    delete(asset_id)

    return



#!/usr/bin/python3

"""Script to rename the kernel of a notebook to a specific name."""

import argparse
import json
from pathlib import Path

from colorama import Fore, init

init()

# init parser
parser = argparse.ArgumentParser(description=__doc__, usage="entry_point")

parser.add_argument("file", help="UI file path")

# Add a optional argument
parser.add_argument("-t", "--test", help="Either test or not (production)", action="store_true")


def main() -> None:
    """Launch the venv creation process."""
    # read arguments (there should be none)
    args = parser.parse_args()

    ui_file = args.file
    test = args.test

    # welcome the user
    print(f"{Fore.YELLOW}UI renaming process{Fore.RESET}")

    # check that the local folder is a module folder
    ui_file = Path.cwd() / ui_file
    if not ui_file.is_file():
        raise Exception(f"{Fore.RED}This is not a module folder.")

    entry_point = Path.cwd() / ui_file

    # create the kernel from venv
    prefix_name = "test-" if test else "venv-"
    prefix_display = "(test) test-" if test else " (venv) "

    name = f"{prefix_name}{Path.cwd().name}"
    display_name = f"{prefix_display}{Path.cwd().name}"

    # change the kernel of the entrypoint to use this one instead
    with entry_point.open() as f:
        data = json.load(f)

    data["metadata"]["kernelspec"]["display_name"] = display_name
    data["metadata"]["kernelspec"]["name"] = name

    entry_point.write_text(json.dumps(data, indent=1))

    # display last message to the end user
    print(
        f'{Fore.GREEN}The python kernel of {args.file} has been updated"{display_name}".{Fore.RESET}'
    )

    return


if __name__ == "__main__":
    main()



"""All the helper function of sepal-ui."""

import configparser
import json
import math
import os
import random
import re
import string
import warnings
from pathlib import Path
from typing import Any, Sequence, Tuple, Union
from urllib.parse import urlparse

import ee
import ipyvuetify as v
import requests
import tomli
from anyascii import anyascii
from deprecated.sphinx import deprecated, versionadded
from matplotlib import colors as c

import sepal_ui
from sepal_ui.conf import config, config_file
from sepal_ui.message import ms
from sepal_ui.scripts import decorator as sd
from sepal_ui.scripts.warning import SepalWarning

# Types
Pathlike = Union[str, Path]


def hide_component(widget: v.VuetifyWidget) -> v.VuetifyWidget:
    """Hide a vuetify based component.

    Args:
        widget: the widget to hide

    Returns:
        the widget
    """
    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.hide()

    elif "d-none" not in str(widget.class_):
        widget.class_ = str(widget.class_).strip() + " d-none"

    return widget


def show_component(widget: v.VuetifyWidget) -> v.VuetifyWidget:
    """Show a vuetify based component.

    Args:
        widget: the widget to hide

    Returns:
        the widget
    """
    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.show()

    elif "d-none" in str(widget.class_):
        widget.class_ = widget.class_.replace("d-none", "")

    return widget


def create_download_link(pathname: Pathlike) -> str:
    """Create a clickable link to download the pathname target.

    Args:
        pathname: the pathname th download

    Returns:
        the download link
    """
    # return the link if it's an absolute url
    if isinstance(pathname, str) and bool(urlparse(str(pathname)).netloc):
        return pathname

    # create a downloadable link from the jupyter node
    pathname = Path(pathname)
    try:
        download_path = pathname.relative_to(Path.home())
    except ValueError:
        download_path = pathname

    # I want to use the ipyurl lib to guess the url of the Jupyter server on the fly
    # but I don't really understand how it works
    # so here is an ugly fix only compatible with SEPAL
    link = f"https://sepal.io/api/sandbox/jupyter/files/{download_path}"

    return link


def random_string(string_length: int = 3) -> str:
    """Generates a random string of fixed length.

    Args:
        string_length: Fixed length. Defaults to 3.

    Returns:
        A random string
    """
    letters = string.ascii_lowercase

    return "".join(random.choice(letters) for i in range(string_length))


def get_file_size(filename: Pathlike) -> str:
    """Get the file size as string of 2 digit in the adapted scale (B, KB, MB....).

    Args:
        filename: the path to the file to measure

    Returns:
        the file size in a readable humanly readable
    """
    file_size = Path(filename).stat().st_size

    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(file_size, 1024))) if file_size > 0 else 0
    s = file_size / (1024**i)

    return "{:.1f} {}".format(s, size_name[i])


def init_ee() -> None:
    r"""Initialize earth engine according using a token.

    THe environment used to run the tests need to have a EARTHENGINE_TOKEN variable.
    The content of this variable must be the copy of a personal credential file that you can find on your local computer if you already run the earth engine command line tool. See the usage question for a github action example.

    - Windows: ``C:\Users\USERNAME\\.config\\earthengine\\credentials``
    - Linux: ``/home/USERNAME/.config/earthengine/credentials``
    - MacOS: ``/Users/USERNAME/.config/earthengine/credentials``

    Note:
        As all init method of pytest-gee, this method will fallback to a regular ``ee.Initialize()`` if the environment variable is not found e.g. on your local computer.
    """
    if not ee.data._credentials:
        credential_folder_path = Path.home() / ".config" / "earthengine"
        credential_file_path = credential_folder_path / "credentials"

        if "EARTHENGINE_TOKEN" in os.environ and not credential_file_path.exists():

            # write the token to the appropriate folder
            ee_token = os.environ["EARTHENGINE_TOKEN"]
            credential_folder_path.mkdir(parents=True, exist_ok=True)
            credential_file_path.write_text(ee_token)

        # Extract the project name from credentials
        _credentials = json.loads(credential_file_path.read_text())
        project_id = _credentials.get("project_id", _credentials.get("project", None))

        if not project_id:
            raise NameError(
                "The project name cannot be detected. "
                "Please set it using `earthengine set_project project_name`."
            )

        # Check if we are using a google service account
        if _credentials.get("type") == "service_account":
            ee_user = _credentials.get("client_email")
            credentials = ee.ServiceAccountCredentials(ee_user, str(credential_file_path))
            ee.Initialize(credentials=credentials)
            ee.data._cloud_api_user_project = project_id
            return

        # if the user is in local development the authentication should
        # already be available
        ee.Initialize(project=project_id)


def normalize_str(msg: str, folder: bool = True) -> str:
    """Normalize an str to make it compatible with file naming (no spaces, special chars ...etc).

    Params:
        msg: the string to sanitise
        folder: if the name will be used for folder naming or for display. if display, <'> and < > characters will be kept

    Returns:
        the modified str
    """
    regex = "[^a-zA-Z\d\-\_]" if folder else "[^a-zA-Z\d\-\_\ ']"

    return re.sub(regex, "_", anyascii(msg))


def to_colors(in_color: Union[str, Sequence], out_type: str = "hex") -> Union[str, tuple]:
    """Transform any color type into a color in the specified output format.

    Available format: [hex]

    Args:
        in_color: It can be a string (e.g., 'red', '#ffff00', 'ffff00') or RGB tuple (e.g., (255, 127, 0)).
        out_type: the type of the output color from ['hex']. default to 'hex'

    Returns:
        The color in the specified format. default to black.
    """
    # list of the color function used for the translation
    c_func = {"hex": c.to_hex}
    transform = c_func[out_type]

    out_color = "#000000"  # default black color

    if isinstance(in_color, tuple) and len(in_color) == 3:
        # rescale color if necessary
        if all(isinstance(item, int) for item in in_color):
            in_color = [c / 255.0 for c in in_color]

        return transform(in_color)

    else:
        # try to guess the color system
        try:
            return transform(in_color)
        except Exception:
            pass

        # try again by adding an extra # (GEE handle hex codes without #)
        try:
            return transform(f"#{in_color}")
        except Exception:
            pass

    return transform(out_color)


def next_string(string: str) -> str:
    """Create a string followed by an underscore and a consecutive number.

    Args:
        string: the initial string

    Returns:
        the incremented string
    """
    # if the string is already numbered the last digit is separated from the rest of the string by an "_"
    split = string.split("_")
    end = split[-1]

    if end.isdigit():
        string = "_".join(split[:-1]) + f"_{int(end)+1}"
    else:
        string += "_1"

    return string


def set_config(key: str, value: str, section: str = "sepal-ui") -> None:
    """Set a config variable.

    Set the provided value to the given key for the given section in the sepal-ui config
    file.

    Args:
        key: key configuration name
        value: value to be referenced by the configuration key
        section: configuration section, defaults to sepal-ui.
    """
    # set the section if needed
    if "sepal-ui" not in config.sections():
        config.add_section(section)

    # set the value
    config.set("sepal-ui", key, value)

    # save back the file
    config.write(config_file.open("w"))

    return


@deprecated(version="2.9.1", reason="This function will be removed in favor of set_config()")
def set_config_locale(locale: str) -> None:
    """Set the provided local in the sepal-ui config file.

    Args:
        locale: a locale name in IETF BCP 47 (no verifications are performed)
    """
    return set_config("locale", locale)


@deprecated(version="2.9.1", reason="This function will be removed in favor of set_config()")
def set_config_theme(theme: str) -> None:
    """Set the provided theme in the sepal-ui config file.

    Args:
        theme: a theme name (currently supporting "dark" and "light")
    """
    return set_config("theme", theme)


@versionadded(version="2.7.1")
def set_type(color: str) -> str:
    r"""Return a pre-defined material colors based on the requested type\_ parameter.

    If the parameter is not a predefined color, fallback to "info" and will raise a warning. the colors can only be selected from ["primary", "secondary", "accent", "error", "info", "success", "warning", "anchor"].

    Args:
        color: the requested color

    Returns:
        a pre-defined material color

    """
    from sepal_ui.frontend.styles import TYPES

    if color not in TYPES:
        warnings.warn(
            f'the selected color "{color}" is not a pre-defined material color. It should be one from [{", ".join(TYPES)}]',
            SepalWarning,
        )
        color = TYPES[0]

    return color


@versionadded(version="2.8.0")
def geojson_to_ee(
    geo_json: dict, geodesic: bool = False, encoding: str = "utf-8"
) -> ee.FeatureCollection:
    """Transform a geojson object into a featureCollection or a Geometry.

    No sanity check is performed on the initial geo_json. It must respect the
    `__geo_interface__ <https://gist.github.com/sgillies/2217756>`__.

    Args:
        geo_json: a geo_json dictionary
        geodesic: Whether line segments should be interpreted as spherical geodesics. If false, indicates that line segments should be interpreted as planar lines in the specified CRS. If absent, defaults to True if the CRS is geographic (including the default EPSG:4326), or to False if the CRS is projected. Defaults to False.
        encoding: The encoding of characters. Defaults to "utf-8".

    Returns:
        the created featurecollection
    """
    # from a featureCollection
    if geo_json["type"] == "FeatureCollection":
        for feature in geo_json["features"]:
            if feature["geometry"]["type"] != "Point":
                feature["geometry"]["geodesic"] = geodesic
        features = ee.FeatureCollection(geo_json)
        return features

    # from a single feature
    elif geo_json["type"] == "Feature":
        geom = None
        # Checks whether it is a point
        if geo_json["geometry"]["type"] == "Point":
            coordinates = geo_json["geometry"]["coordinates"]
            longitude = coordinates[0]
            latitude = coordinates[1]
            geom = ee.Geometry.Point(longitude, latitude)
        # for every other geometry simply create a geometry
        else:
            geom = ee.Geometry(geo_json["geometry"], "", geodesic)

        return geom

    # some error handling because we are fancy
    else:
        raise ValueError("Could not convert the geojson to ee.Geometry()")

    return


def check_input(input_: Any, msg: str = ms.utils.check_input.error) -> bool:
    r"""Check if the inpupt value is initialized.

    If not raise an error, else return True.

    Args:
        input\_: the input to check
        msg: the message to display if the input is not set

    Return:
        check if the value is initialized
    """
    # by the default the variable is considered valid
    init = True

    # check the collection type that are the only one supporting the len method
    try:
        init = False if len(input_) == 0 else init
    except Exception:
        init = False if input_ is None else init

    if init is False:
        raise ValueError(msg)

    return init


def get_app_version(repo_folder: Pathlike = Path.cwd()) -> str:
    """Get the current version of the a github project using the pyproject.toml file in the root.

    Returns:
        the version of the repository
    """
    # get the path to the pyproject.toml file
    pyproject_path = repo_folder / "pyproject.toml"

    # check if the file exist
    if pyproject_path.exists():
        # read the file using tomli
        pyproject = tomli.loads(pyproject_path.read_text())

        # get the version
        return pyproject.get("project", {}).get("version", None)

    return None


def get_repo_info(repo_folder: Pathlike = Path.cwd()) -> Tuple[str, str]:
    """Get the repository name and owner from the git config file."""
    config = configparser.ConfigParser()
    git_config_path = Path(repo_folder) / ".git/config"
    config.read(git_config_path)

    try:
        remote_url = config.get('remote "origin"', "url")
    except (configparser.NoSectionError, configparser.NoOptionError):
        return "", ""

    # Check if URL is likely SSH
    if "git@" in remote_url:
        match = re.search(r":(.*?)/(.*?)(?:\.git)?$", remote_url)

    # Assume URL is HTTPS otherwise
    else:
        match = re.search(r"github\.com/(.*?)/(.*?)(?:\.git)?$", remote_url)

    if match:
        return match.groups()

    else:
        return "", ""


def get_changelog(repo_folder: Pathlike = Path.cwd()) -> str:
    """Check if the repository contains a changelog file and/or a remote release and return its content.

    Returns:
        str: the content of the release and/or changelog file
    """
    changelog_text, release_text = "", ""
    repo_owner, repo_name = get_repo_info(repo_folder)

    release_url = f"https://api.github.com/repos/{repo_owner}/{repo_name}/releases/latest"

    response = requests.get(release_url)
    if all([repo_owner, repo_name]) and response.status_code == 200:
        response_json = response.json()
        name = response_json.get("name")

        if name == f"v_{get_app_version()}":
            release_text = response_json.get("body")

            url_pattern = r"https://github\.com/[^ ]+/pull/\d+"

            # Replace URLs with <a> tags
            def wrap_url_in_a_tag(match):
                url = match.group(0)
                return f'<a href="{url}">{url}</a>'

            release_text = re.sub(url_pattern, wrap_url_in_a_tag, release_text)

    # get the path to the pyproject.toml file
    changelog_path = Path(repo_folder) / "CHANGELOG.md"

    # check if the file exist
    if changelog_path.exists():
        changelog_text = changelog_path.read_text()

    return release_text, changelog_text


################################################################################
# the soon to be deprecated decorators
#

# fmt: off
catch_errors = deprecated(version='3.0', reason="use sepal_ui.scripts.decorator.catch_errors instead")(sd.catch_errors)
need_ee = deprecated(version='3.0', reason="use sepal_ui.scripts.decorator.need_ee instead")(sd.need_ee)
loading_button = deprecated(version='3.0', reason="use sepal_ui.scripts.decorator.need_ee instead")(sd.loading_button)
switch = deprecated(version='3.0', reason="use sepal_ui.scripts.decorator.switch instead")(sd.switch)
# fmt: on



"""Initialization of the ``Translator`` used in the sepal-ui.

Can be accessed as:

.. code-block::

    from sepal_ui.message import ms
"""

from pathlib import Path

from sepal_ui.translator import Translator

ms = Translator(Path(__file__).parent)
